/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// local ALF includes
#include "alf_misc_ppu.h"

//********************************************************
// APIs
//********************************************************

//find a string contains pattern
static int 
_alf_find_string(char* str, char* pattern)
{
  while(*str == ' ') str++;

  if (strlen(str) < strlen(pattern))
  {
    return 0;
  }

  return (strncmp(str, pattern, strlen(pattern)) == 0);
}

/**
 * Get system physical memory size
 *
 * Return Values:
 *           >=0  The system physical memory size by kB.
 *            <0  Errors.
 */
long long 
_alf_get_system_pmem_size()
{
  char filename[] = "/proc/meminfo";
  char token[] = "MemTotal:";
  FILE* fp;
  char line[256];
  long long mem_size = -1;

  fp = fopen(filename, "r");
  if (fp == NULL)
  {
    return -errno;
  }

  while ((fgets(line, 256, fp)) != NULL)
  {
    if (_alf_find_string(line, token))
    {
       sscanf(line, "MemTotal: %lld kB", &mem_size);
       break;
    }
  }

  fclose(fp);

  return mem_size; 
}




