/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#ifndef _REALLOC_ALIGN_H_
#define _REALLOC_ALIGN_H_	1

#include <stdlib.h>

/* Function
 *
 *	void * realloc_align(void *ptr, size_t size, unsigned int log2_align)
 *
 * Description
 * 	The realloc_align routine re-allocates a memory buffer pointed to
 *	by ptr to <size> bytes aligned to the power of 2 alignment specified 
 *	by <log2_align>. For example, realloc_align(ptr, 4096, 7) will re-allocate
 *	the buffer pointed to by <ptr> to 4096 bytes aligned on a 128 byte boundary.
 *
 *	The input buffer pointed to by <ptr> must have been previously allocated
 *	using the malloc_align routine or calloc_align routine. The contents of the 
 *	buffer are only preserved if the <align> is the same as alignment of the original
 *	buffer allocation.
 */

static __inline void * _realloc_align(void *ptr, size_t size, unsigned int log2_align)
{
  void *ret;
  char *real;
  void *real_ptr;
  unsigned long offset;
  unsigned long align;

  real_ptr = *((void **)(ptr)-1);
  
  align = 1 << log2_align;
  real = (char *)realloc(real_ptr, size + sizeof(void *) + (align-1));
  if (real) {
    offset = (align - (unsigned long)(real + sizeof(void *))) & (align-1);
    ret = (void *)(real + sizeof(void *)) + offset;
    *((void **)(ret)-1) = (void *)(real);
  } else {
    ret = (void *)(real);
  }
  return (ret);
}

#endif /* _REALLOC_ALIGN_H_ */
